"""Core hashing functions."""

from __future__ import annotations

import collections
import importlib.metadata
import subprocess
from typing import TYPE_CHECKING

import xxhash

import qik.shell

if TYPE_CHECKING:
    import qik.dep as run_deps


def globs(*vals: run_deps.Glob | str) -> str:
    """Compute a hash string of glob patterns.

    Use git ls-files, tracking both added and modified file hashes.
    """
    if not vals:
        return ""

    fmt = "--format '%(path)\t%(objectname)'"
    # Create a pattern string for git ls-files. Ensure there are no duplicates and
    # that we sort globs for a consistent hash
    pattern_str = " ".join(sorted({f"'{glob}'" for glob in vals}))
    git_ls_lines = qik.shell.exec(f"git ls-files -cm {fmt} {pattern_str}", check=True, lines=True)
    git_ls_lines_split = [line.split("\t", 1) for line in git_ls_lines]
    hashes = dict(git_ls_lines_split)

    # Files that are modified appear twice. Manually compute their hashes
    path_counts = collections.Counter(line[0] for line in git_ls_lines_split)
    modified = [path for path, count in path_counts.items() if count > 1]
    if modified:
        try:
            modified_hashes_lines = qik.shell.exec(
                f"git ls-files --format '%(path)' {pattern_str} -m | xargs git hash-object",
                check=True,
                lines=True,
            )
        except subprocess.CalledProcessError:
            # If there are issues with the first command, it likely means a file does
            # not exist. Do the suboptimal strategy here, piping individual files to
            # `git hash-object` while printing zeroes for files that no longer exist.
            cmd = f"""
                git ls-files --format '%(path)' {pattern_str} -m | while IFS= read -r file; do
                    if [ -f "$file" ]; then
                        git hash-object "$file"
                    else
                        echo "0000000000000000000000000000000000000000"
                    fi
                done
            """
            modified_hashes_lines = qik.shell.exec(cmd, check=True, lines=True)

        for i, name in enumerate(modified):
            hashes[name] = modified_hashes_lines[i]

    return xxhash.xxh128_hexdigest("".join(f"{name}{hash}" for name, hash in hashes.items()))


def dists(*vals: str) -> str:
    return xxhash.xxh128_hexdigest(
        "".join(f"{dist}{importlib.metadata.version(dist)}" for dist in sorted(vals))
    )


def strs(*vals: str) -> str:
    return xxhash.xxh128_hexdigest("".join(sorted(vals)))


def val(input: str | bytes) -> str:
    return xxhash.xxh128_hexdigest(input)



"""Qik project and configuration loading."""

from __future__ import annotations

import functools
import importlib.util
import os.path
import pathlib
import sys
from typing import Any, Literal, TypeAlias, TypeVar

import msgspec.structs
import msgspec.toml

import qik.unset

CtxNamespace: TypeAlias = Literal["qik", "project", "modules", "plugins"]
VarType: TypeAlias = str | bool | int
VAR_T = TypeVar("VAR_T", str, bool, int, None)
CacheBackend: TypeAlias = Literal["s3"]
CacheWhen: TypeAlias = Literal["success", "failed", "finished"]
CacheStatus: TypeAlias = Literal["warm", "code"]


class Base(
    msgspec.Struct,
    frozen=True,
    omit_defaults=True,
    forbid_unknown_fields=True,
    rename="kebab",
    dict=True,
):
    pass


class BaseDep(Base, frozen=True):
    pass


class GlobDep(BaseDep, tag="glob", frozen=True):
    pattern: str


class ConstDep(BaseDep, tag="const", frozen=True):
    val: str


class ValDep(BaseDep, tag="val", frozen=True):
    key: str
    file: str


class CmdDep(BaseDep, tag="command", frozen=True):
    name: str
    strict: bool = False
    isolated: bool | qik.unset.UnsetType = qik.unset.UNSET


class DistDep(BaseDep, tag="dist", frozen=True):
    name: str


class ModuleDep(BaseDep, tag="module", frozen=True):
    name: str


DepType: TypeAlias = str | GlobDep | CmdDep | DistDep | ModuleDep | ConstDep


class CmdConf(Base, frozen=True):
    exec: str = ""
    deps: list[DepType] = []
    artifacts: list[str] = []
    cache: str | qik.unset.UnsetType = qik.unset.UNSET
    cache_when: CacheWhen | qik.unset.UnsetType = qik.unset.UNSET
    factory: str = ""


class Var(Base, frozen=True):
    name: str
    type: Literal["str", "int", "bool"] = "str"
    default: VarType | qik.unset.UnsetType = qik.unset.UNSET
    required: bool = True

    def __post_init__(self):
        if not self.required and self.default is qik.unset.UNSET:
            msgspec.structs.force_setattr(self, "default", None)

    @property
    def py_type(self) -> type[VarType]:
        if self.required:
            return __builtins__[self.type]
        else:
            return __builtins__[self.type] | None


class ModuleConf(Base, frozen=True):
    vars: list[str | Var] = []
    commands: dict[str, CmdConf] = {}

    @functools.cached_property
    def vars_dict(self) -> dict[str, Var]:
        return dict((v, Var(v)) if isinstance(v, str) else (v.name, v) for v in self.vars)


class ModulePath(Base, frozen=True):
    name: str
    path: str

    @functools.cached_property
    def dir(self) -> pathlib.Path:
        return root() / self.path.replace(".", os.path.sep).replace("/", os.path.sep)

    @functools.cached_property
    def imp(self) -> str:
        return self.path.replace("/", ".")

    @functools.cached_property
    def conf(self) -> ModuleConf:
        try:
            return msgspec.toml.decode(
                (self.dir / "qik.toml").read_bytes(),
                type=ModuleConf,
            )
        except FileNotFoundError:
            return ModuleConf()


class PluginPath(ModulePath, frozen=True):
    @functools.cached_property
    def dir(self) -> pathlib.Path:
        spec = importlib.util.find_spec(self.path)
        if not spec or not spec.origin:
            raise RuntimeError(f'Could not import plugin "{self.name}"')

        return pathlib.Path(spec.origin).parent


class Env(Base, frozen=True):
    lock_file: str | list[str] = []


class Graph(Base, frozen=True):
    include_type_checking: bool = True
    include_dists: bool = True


class Cache(Base, frozen=True, tag_field="type"):
    pass


class S3Cache(Cache, frozen=True, tag="s3"):
    bucket: str
    prefix: str = ""
    aws_access_key_id: str | None = None
    aws_secret_access_key: str | None = None
    aws_session_token: str | None = None
    region_name: str | None = None
    endpoint_url: str | None = None


class ProjectConf(ModuleConf, frozen=True):
    modules: list[str | ModulePath] = []
    plugins: list[str | PluginPath] = []
    deps: list[DepType] = []
    ctx: dict[str, dict[CtxNamespace, dict[str, Any]]] = {}
    venvs: dict[str, Env] = {}
    caches: dict[str, S3Cache] = {}
    graph: Graph = msgspec.field(default_factory=Graph)

    @functools.cached_property
    def modules_by_name(self) -> dict[str, ModulePath]:
        return dict(
            (m, ModulePath(m, m)) if isinstance(m, str) else (m.name, m) for m in self.modules
        )

    @functools.cached_property
    def modules_by_path(self) -> dict[str, ModulePath]:
        return {m.path: m for m in self.modules_by_name.values()}

    @functools.cached_property
    def plugins_by_name(self) -> dict[str, PluginPath]:
        return dict(
            (p, PluginPath(p, p)) if isinstance(p, str) else (p.name, p) for p in self.plugins
        )

    @functools.cached_property
    def plugins_by_path(self) -> dict[str, PluginPath]:
        return {p.path: p for p in self.plugins_by_name.values()}


class PyprojectTool(msgspec.Struct):
    qik: ProjectConf | None = None


class Pyproject(msgspec.Struct, omit_defaults=True):
    tool: PyprojectTool | None = None


@functools.cache
def _project_conf() -> tuple[ProjectConf, pathlib.Path]:
    """Return the project configuration and file."""
    cwd = pathlib.Path.cwd()
    qik_toml: pathlib.Path | None = None

    for directory in (cwd, *cwd.parents):
        if (directory / "qik.toml").is_file():
            qik_toml = directory / "qik.toml"

        if (
            has_pyproject := (directory / "pyproject.toml").is_file()
            or (directory / ".git").is_dir()
        ):
            if qik_toml and qik_toml.parent == directory:
                # qik.toml is at the root. Use qik.toml
                return msgspec.toml.decode(qik_toml.read_bytes(), type=ProjectConf), qik_toml
            elif has_pyproject:
                location = directory / "pyproject.toml"
                pyproject = msgspec.toml.decode(location.read_bytes(), type=Pyproject)
                if pyproject.tool and pyproject.tool.qik:
                    return pyproject.tool.qik, location
                elif qik_toml:  # qik.toml was found but not at root
                    return msgspec.toml.decode(qik_toml.read_bytes(), type=ProjectConf), qik_toml

            break

    raise RuntimeError("Could not locate qik configuration in qik.toml or pyproject.toml.")


@functools.cache
def project() -> ProjectConf:
    """Get project-level configuration."""
    sys.path.append(str(root()))
    return _project_conf()[0]


@functools.cache
def path_to_name(path: str) -> str:
    """Given a full path, return the module or plugin name."""
    proj = project()
    if path in proj.modules_by_path:
        return proj.modules_by_path[path].name
    elif path in proj.plugins_by_path:
        return proj.plugins_by_path[path].name
    else:
        raise KeyError(f'No configured module or plugin with path "{path}".')


@functools.cache
def module(name: str) -> ModuleConf:
    """Get module configuration."""
    proj = project()
    if name not in proj.modules_by_name:
        raise KeyError(f'Module "{name}" not configured in {location().name}.')

    return proj.modules_by_name[name].conf


def plugin(name: str) -> ModuleConf:
    """Get plugin configuration."""
    proj = project()
    if name not in proj.plugins_by_name:
        raise KeyError(f'Plugin "{name}" not configured in {location().name}.')

    return proj.plugins_by_name[name].conf


@functools.cache
def get(name: str | None = None) -> ModuleConf:
    """Get configuration for a given module, plugin, or project."""
    if not name:
        return project()
    else:
        try:
            return module(name)
        except KeyError:
            try:
                return plugin(name)
            except KeyError:
                raise KeyError(f'Module or plugin "{name}" not configured in {location().name}.')  # noqa: B904


@functools.cache
def uri_parts(uri: str) -> tuple[str | None, str]:
    """Return the module and name of a URI."""
    return (None, uri) if "." not in uri else tuple(uri.rsplit(".", 1))  # type: ignore


@functools.cache
def command(uri: str) -> CmdConf:
    """Get configuration for a command."""
    module, name = uri_parts(uri)
    conf = get(module)
    if name not in conf.commands:
        raise KeyError(f'Command "{uri}" not configured.')

    return conf.commands[name]


@functools.cache
def root() -> pathlib.Path:
    """Get the absolute root project directory."""
    return _project_conf()[1].parent


@functools.cache
def priv_work_dir(relative: bool = False) -> pathlib.Path:
    """Get the private work directory."""
    return root() / "._qik" if not relative else pathlib.Path("._qik")


@functools.cache
def pub_work_dir(relative: bool = False) -> pathlib.Path:
    """Get the public work directory."""
    return root() / ".qik" if not relative else pathlib.Path(".qik")


@functools.cache
def location() -> pathlib.Path:
    """Get the root configuration file."""
    return _project_conf()[1]



from __future__ import annotations

import argparse
import os
import sys

import qik.ctx
import qik.runner
import qik.unset


def main() -> None:
    """The main entrypoint into the CLI."""
    parser = argparse.ArgumentParser()

    parser.add_argument("commands", help="Command name(s)", nargs="*")
    parser.add_argument(
        "-m",
        "--module",
        help="Set module name(s).",
        action="append",
        dest="modules",
    )
    parser.add_argument(
        "--watch", action="store_true", help="Watch for changes.", default=qik.unset.UNSET
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Don't read cache.",
        default=qik.unset.UNSET,
    )
    parser.add_argument("--cache", help="Set default cache.", default=qik.unset.UNSET)
    parser.add_argument(
        "--cache-when",
        help="Cache results by event type.",
        default=qik.unset.UNSET,
        choices=["success", "failed", "finished"],
    )
    parser.add_argument(
        "-n", "--workers", help="Number of workers.", type=int, default=qik.unset.UNSET
    )
    parser.add_argument(
        "--isolated",
        action="store_true",
        help="Don't run dependent commands.",
        default=qik.unset.UNSET,
    )
    parser.add_argument(
        "--ls", action="store_true", help="List selected commands.", default=qik.unset.UNSET
    )
    parser.add_argument(
        "--fail",
        action="store_true",
        help="Fail if any commands are selected.",
        default=qik.unset.UNSET,
    )
    parser.add_argument("-p", "--profile", help="The context profile to use.")
    parser.add_argument("--since", help="Filter since git SHA.", default=qik.unset.UNSET)
    parser.add_argument(
        "--cache-type",
        help="Filter by cache type.",
        action="append",
        dest="cache_types",
    )
    parser.add_argument(
        "--cache-status",
        help="Filter by cache status.",
        default=qik.unset.UNSET,
        choices=["warm", "cold"],
    )
    parser.add_argument(
        "-v",
        "--verbosity",
        nargs="?",
        const=2,
        default=qik.unset.UNSET,
        type=int,
        help="Set verbosity (1 by default, 2 if -v is present, or specify level)",
    )

    args = parser.parse_args()
    profile = args.profile or os.environ.get("QIK__PROFILE", "default")
    with (
        qik.ctx.set_profile(profile),
        qik.ctx.set_vars(
            "qik",
            watch=args.watch,
            force=args.force,
            cache=args.cache,
            isolated=args.isolated,
            ls=args.ls,
            workers=args.workers,
            since=args.since,
            fail=args.fail,
            cache_status=args.cache_status,
            cache_when=args.cache_when,
            verbosity=args.verbosity,
            commands=args.commands or qik.unset.UNSET,
            modules=args.modules or qik.unset.UNSET,
            cache_types=args.cache_types or qik.unset.UNSET,
        ),
    ):
        res = qik.runner.exec()
        qik_ctx = qik.ctx.module("qik")

        if qik_ctx.ls:
            for cmd in sorted(res, key=lambda k: k.name):
                print(cmd.name)

        if res and qik_ctx.fail:
            sys.exit(1)



from __future__ import annotations

import functools
import importlib.metadata
import os
import pathlib
import sysconfig
import threading
from typing import TYPE_CHECKING, Iterator

import msgspec

import qik.conf
import qik.dep
import qik.file
import qik.graph.core
import qik.hash
import qik.runnable

if TYPE_CHECKING:
    import grimp.exceptions as grimp_exceptions
else:
    import qik.lazy

    grimp_exceptions = qik.lazy.module("grimp.exceptions")


class GraphConfDep(qik.dep.Val, frozen=True, dict=True):
    val: str = "graph"
    file: str = ""  # The file is dynamic based on config location

    @functools.cached_property
    def watch(self) -> list[str]:
        return [qik.conf.location().name]

    # TODO: break this cache on runner invocations
    @functools.cached_property
    def vals(self) -> list[str]:
        return [str(msgspec.json.encode(qik.conf.project().graph))]


@functools.cache
def _graph_conf_dep() -> qik.dep.Val:
    """Serialize the graph config so it can be included as a dependency."""
    return GraphConfDep()


def build_cmd(runnable: qik.runnable.Runnable) -> tuple[int, str]:
    """Build the graph."""
    try:
        graph = qik.graph.core.build()
    except grimp_exceptions.SourceSyntaxError as exc:
        return 1, f"Cannot build import graph - {exc}"

    path = graph_path()

    serialized = msgspec.json.encode(graph)
    qik.file.write(path, serialized)
    return 0, ""


class PackagesDistributions(msgspec.Struct):
    venv_hash: str
    packages_distributions: dict[str, list[str]]


@functools.lru_cache(maxsize=1)
def _packages_distributions(venv_hash: str) -> dict[str, list[str]]:
    cache_path = qik.conf.priv_work_dir() / "graph" / "packages_distributions.json"
    try:
        cached_val = msgspec.json.decode(cache_path.read_bytes(), type=PackagesDistributions)
        if cached_val.venv_hash == venv_hash:
            return cached_val.packages_distributions
    except FileNotFoundError:
        pass

    cached_val = PackagesDistributions(
        venv_hash=venv_hash, packages_distributions=importlib.metadata.packages_distributions()
    )
    qik.file.write(cache_path, msgspec.json.encode(cached_val))

    return cached_val.packages_distributions


_PACKAGES_DISTRIBUTIONS_LOCK = threading.Lock()


def packages_distributions() -> dict[str, list[str]]:
    """Obtain a mapping of modules to their associated dists.

    This is an expensive command, so use an underlying cache when possible.
    """
    venv_hash = qik.hash.strs(*os.listdir(sysconfig.get_path("purelib")))
    with _PACKAGES_DISTRIBUTIONS_LOCK:
        return _packages_distributions(venv_hash)


def analyze_cmd(runnable: qik.runnable.Runnable) -> tuple[int, str]:
    if not runnable.module:
        raise AssertionError("Unexpected analyze runnable.")

    graph = load_graph()  # TODO: Use cached runner graph
    # TODO: Better error if the module doesn't exist
    module = qik.conf.project().modules_by_name[runnable.module]
    upstream = graph.upstream_modules(module.imp, idx=False)
    root = graph.modules[graph.modules_idx[module.imp]]
    distributions = packages_distributions()

    def _gen_upstream_globs() -> Iterator[str]:
        for module in [root, *upstream]:
            if module.is_internal:
                path = module.imp.replace(".", "/")
                yield f"{path}/**.py"
                yield f"{path}.py"

    def _gen_upstream_dists() -> Iterator[str]:
        for module in upstream:
            if not module.is_internal:
                yield from distributions[module.imp]

    qik.dep.store(
        analysis_path(module.name),
        globs=sorted(_gen_upstream_globs()),
        dists=sorted(_gen_upstream_dists()),
    )
    return 0, ""


def build_cmd_factory(cmd: str, conf: qik.conf.CmdConf) -> dict[str, qik.runnable.Runnable]:
    build_graph_cmd_name = build_cmd_name()
    runnable = qik.runnable.Runnable(
        name=build_graph_cmd_name,
        cmd=build_graph_cmd_name,
        val="qik.graph.cmd.build_cmd",
        shell=False,
        deps=[qik.dep.Glob("**.py"), _graph_conf_dep(), *qik.dep.project_deps()],
        artifacts=[str(graph_path())],
        cache="repo",
    )
    return {runnable.name: runnable}


def analyze_cmd_factory(cmd: str, conf: qik.conf.CmdConf) -> dict[str, qik.runnable.Runnable]:
    cmd_name = analyze_cmd_name()
    build_graph_dep = qik.dep.Cmd(build_cmd_name())

    runnables = [
        qik.runnable.Runnable(
            name=f"{cmd_name}@{module.name}",
            cmd=cmd_name,
            val="qik.graph.cmd.analyze_cmd",
            shell=False,
            deps=[
                build_graph_dep,
                qik.dep.Load(
                    str(analysis_path(module.name)),
                    default=["**.py"],
                ),
                _graph_conf_dep(),
                *qik.dep.project_deps(),
            ],
            artifacts=[str(analysis_path(module.name))],
            module=module.name,
            cache="repo",
        )
        for module in qik.conf.project().modules_by_name.values()
    ]

    return {runnable.name: runnable for runnable in runnables}


@functools.cache
def build_cmd_name() -> str:
    graph_plugin_name = qik.conf.path_to_name("qik.graph")
    return f"{graph_plugin_name}.build"


@functools.cache
def analyze_cmd_name() -> str:
    graph_plugin_name = qik.conf.path_to_name("qik.graph")
    return f"{graph_plugin_name}.analyze"


@functools.cache
def graph_path() -> pathlib.Path:
    return qik.conf.pub_work_dir(relative=True) / "artifacts" / build_cmd_name() / "graph.json"


@functools.cache
def analysis_path(module: str) -> pathlib.Path:
    return (
        qik.conf.pub_work_dir(relative=True)
        / "artifacts"
        / analyze_cmd_name()
        / f"analzye.{module}.json"
    )


def load_graph() -> qik.graph.core.Graph:
    """Load the graph."""
    return msgspec.json.decode(graph_path().read_bytes(), type=qik.graph.core.Graph)



from __future__ import annotations

import functools
import pkgutil
from typing import TYPE_CHECKING, Iterator

import msgspec

import qik.conf
import qik.unset

if TYPE_CHECKING:
    from qik.runnable import Runnable


class Cmd(msgspec.Struct, frozen=True):
    name: str
    runnables: dict[str, Runnable]


@functools.cache
def load(cmd: str) -> Cmd:
    """Load runnables for a command."""
    cmd_conf = qik.conf.command(cmd)
    runnables = pkgutil.resolve_name(cmd_conf.factory or "qik.runnable.factory")(cmd, cmd_conf)
    return Cmd(name=cmd, runnables=runnables)


def ls() -> Iterator[Cmd]:
    """List all commands."""
    proj = qik.conf.project()
    for module in [None, *proj.modules_by_name, *proj.plugins_by_name]:
        conf = qik.conf.get(module)
        for command in conf.commands:
            yield load(f"{module}.{command}") if module else load(command)



from __future__ import annotations

import fnmatch
import functools
import os
import pkgutil
import re
from typing import TYPE_CHECKING, Literal, TypeAlias

import msgspec
from typing_extensions import Self

import qik.cache
import qik.conf
import qik.ctx
import qik.dep
import qik.hash
import qik.shell
import qik.unset

if TYPE_CHECKING:
    Direction: TypeAlias = Literal["up", "down"]
    Edges: TypeAlias = dict[str, set[str]]
    FilterStrategy: TypeAlias = Literal["since", "watch"]

    import qik.cache
    import qik.logger


def _get_exec_env() -> dict[str, str]:
    """Get the environment for runnables."""
    return {
        **os.environ,
        "QIK__CMD": qik.ctx.runnable().cmd,
        "QIK__RUNNABLE": qik.ctx.runnable().name,
        "QIK__WORKER": str(qik.ctx.worker_id()),
    }


def _make_runnable(
    *,
    cmd: str,
    conf: qik.conf.CmdConf,
    module: qik.conf.ModulePath | None = None,
) -> Runnable:
    return Runnable(
        name=f"{cmd}@{module.name}" if module else cmd,
        cmd=cmd,
        val=qik.ctx.format(conf.exec, module=module),
        deps=[
            *(qik.dep.factory(dep, module=module) for dep in conf.deps),
            *qik.dep.project_deps(),
        ],
        module=module.name if module else None,
        artifacts=[qik.ctx.format(artifact) for artifact in conf.artifacts],
        cache=qik.ctx.format(conf.cache),
        cache_when=qik.ctx.format(conf.cache_when),
    )


def factory(cmd: str, conf: qik.conf.CmdConf) -> dict[str, Runnable]:
    if "{module" in conf.exec:
        runnables = (
            _make_runnable(cmd=cmd, conf=conf, module=module)
            for module in qik.conf.project().modules_by_name.values()
        )
    else:
        runnables = [_make_runnable(cmd=cmd, conf=conf)]

    return {runnable.name: runnable for runnable in runnables}


class Result(msgspec.Struct, frozen=True):
    log: str | None
    code: int
    hash: str

    @classmethod
    def from_cache(cls, entry: qik.cache.Entry) -> Self:
        return cls(log=entry.log, code=entry.manifest.code, hash=entry.manifest.hash)


@functools.cache
def _glob_to_regex(glob_pattern: str) -> str:
    """Translate a glob to a regex pattern"""
    return fnmatch.translate(glob_pattern).replace("?s:", "^").replace(r"\Z", "$")


class Runnable(msgspec.Struct, frozen=True, dict=True):
    name: str
    cmd: str
    val: str
    shell: bool = True
    deps: list[qik.dep.BaseDep] = []
    artifacts: list[str] = []
    module: str | None = None
    cache: str | None | qik.unset.UnsetType = qik.unset.UNSET
    cache_when: qik.conf.CacheWhen | qik.unset.UnsetType = qik.unset.UNSET

    @functools.cached_property
    def deps_collection(self) -> qik.dep.Collection:
        return qik.dep.Collection(*self.deps, module=self.module)

    # TODO cache this based on the runner session
    def filter_regex(self, strategy: FilterStrategy) -> re.Pattern | None:
        """Generate the regex used for file-based filtering.

        File-based filtering occurs during --since and --watch.
        """
        globs = self.deps_collection.since if strategy == "since" else self.deps_collection.watch
        files_regex = ")|(".join(_glob_to_regex(glob) for glob in globs)
        return re.compile(f"({files_regex})", re.M) if files_regex else None

    @functools.cached_property
    def spec_hash(self) -> str:
        """Hash the runnable spec."""
        return qik.hash.val(msgspec.json.encode(self))

    def hash(self) -> str:
        """Compute the hash, including the command definitions and deps."""
        return qik.hash.strs(self.spec_hash, self.deps_collection.hash())

    def get_cache_when(self) -> qik.conf.CacheWhen:
        return (
            qik.ctx.module("qik").cache_when
            if self.cache_when is qik.unset.UNSET
            else self.cache_when
        )

    def should_cache(self, code: int) -> bool:
        match self.get_cache_when():
            case "success":
                return code == 0
            case "failed":
                return code != 0
            case "finished":
                return True
            case other:
                raise AssertionError(f'Unexpected cache_when "{other}".')

    def get_cache_backend(self) -> qik.cache.Cache:
        backend = qik.ctx.module("qik").cache if self.cache is qik.unset.UNSET else self.cache
        return qik.cache.load(backend)

    def get_cache_entry(self, artifacts: bool = True) -> qik.cache.Entry | None:
        if not qik.ctx.module("qik").force:
            entry = self.get_cache_backend().get(self, artifacts=artifacts)
            if entry and self.should_cache(entry.manifest.code):
                return entry

    def cache_result(self, result: Result) -> None:
        if self.should_cache(result.code):
            self.get_cache_backend().set(self, result)

    def _uncached_exec(self, logger: qik.logger.Logger) -> Result:
        """Run a command without wrapped caching."""
        if self.shell:
            process = qik.shell.popen(self.val, env=_get_exec_env())
            output = []
            for line in process.stdout if process.stdout is not None else []:
                logger.print(line, runnable=self, event="output")
                output.append(line)

            process.wait()
            code = process.returncode
            log = "".join(output)
        else:
            code, log = pkgutil.resolve_name(self.val)(runnable=self)
            logger.print(log, runnable=self, event="output")

        return Result(log=log, code=code, hash=self.hash())

    def _exec(self) -> Result:
        """Run a command, caching the results."""
        logger = qik.ctx.runner().logger
        cache_entry = self.get_cache_entry()
        print_kwargs = {"cache_entry": cache_entry, "runnable": self}

        if cache_entry:
            # Run cached command
            logger.print(
                msg=f"{self.name} [default][dim]{self.val}",
                emoji="fast-forward_button",
                color="cyan",
                event="start",
                **print_kwargs,
            )
            if cache_entry.log:
                logger.print(cache_entry.log, event="output", **print_kwargs)

            result = Result.from_cache(cache_entry)
        else:
            # Run uncached command
            logger.print(
                msg=f"{self.name} [default][dim]{self.val}",
                emoji="construction",
                color="cyan",
                event="start",
                **print_kwargs,
            )
            result = self._uncached_exec(logger=logger)
            self.cache_result(result)

        if result.code == 0:
            logger.print(
                msg=f"{self.name} [default][dim]{self.val}",
                emoji="white_check_mark",
                color="green",
                event="finish",
                result=result,
                **print_kwargs,
            )
        else:
            logger.print(
                msg=f"{self.name} [default][dim]{self.val}",
                emoji="broken_heart",
                color="red",
                event="finish",
                result=result,
                **print_kwargs,
            )

        return result

    def exec(self) -> Result:
        with qik.ctx.set_runnable(self), qik.ctx.set_worker_id():
            return self._exec()



import functools
import pathlib
import sysconfig

import msgspec

import qik.conf


class Env(msgspec.Struct, frozen=True, dict=True):
    name: str
    lock_file: list[str] = []

    @functools.cached_property
    def dir(self) -> pathlib.Path:
        return pathlib.Path(sysconfig.get_path("purelib"))


@functools.cache
def load(name: str = "default") -> Env:
    """Load an env."""
    proj = qik.conf.project()
    if conf := proj.venvs.get(name):
        lock_file = conf.lock_file if isinstance(conf.lock_file, list) else [conf.lock_file]
        return Env(name=name, lock_file=lock_file)
    elif name == "default":
        return Env(name="default")
    else:
        raise KeyError(f'Venv named "{name}" not configured in qik.venvs.')



from __future__ import annotations

import functools
import pathlib
import re
from typing import TYPE_CHECKING, ClassVar

import msgspec

import qik.cmd
import qik.conf
import qik.ctx
import qik.file
import qik.hash
import qik.unset
import qik.venv

if TYPE_CHECKING:
    import qik.graph.cmd as graph_cmd
    import qik.runnable
else:
    import qik.lazy

    graph_cmd = qik.lazy.module("qik.graph.cmd")


@functools.cache
def _normalize_dist_name(dist: str) -> str:
    dist = dist.lower().strip()
    return re.sub(r"[^a-z0-9]+", "-", dist)


class Runnable(msgspec.Struct, frozen=True):
    """A runnable included by a dependency."""

    name: str
    obj: qik.runnable.Runnable
    strict: bool = False
    isolated: bool | qik.unset.UnsetType = qik.unset.UNSET


class BaseDep(msgspec.Struct, frozen=True, tag=True, dict=True):
    """A base dependency."""

    val: str

    def __str__(self) -> str:
        return self.val

    @functools.cached_property
    def runnables(self) -> list[Runnable]:
        """Return any runnables of the dependency."""
        return []

    @functools.cached_property
    def globs(self) -> list[str]:
        """Return any glob patterns of the dependency."""
        return []

    @functools.cached_property
    def dists(self) -> list[str]:
        """Return any dists of the dependency."""
        return []

    @functools.cached_property
    def vals(self) -> list[str]:
        """Return any vals of the dependency."""
        return []

    @functools.cached_property
    def watch(self) -> list[str]:
        """Return any globs used for --watch option.

        When using --watch, we need to react to events based on file
        system changes. Similar to `since`, we return globs
        used by the watcher to determine changes. The main difference
        between `since` and `watch` is that the watcher has access to
        more than just git files (venv, etc). As a result, we don't
        need to react to as many files in the git repo, such as
        reacting based on a lock file change.
        """
        return self.globs

    @functools.cached_property
    def since(self) -> list[str]:
        """Return any globs used for --since option.

        When using --since, we need to estimate which commands need to
        be re-executed based on git changes. For deps such as
        dists or vals that aren't directly tied to git, we return globs
        that encapsulate those changes (lock files, parent files, etc).
        """
        return self.watch


def factory(conf: qik.conf.BaseDep | str | pathlib.Path, module: str | None = None) -> BaseDep:
    """A factory for creating dependencies from a configuration."""

    def _fmt(val: str) -> str:
        return qik.ctx.format(val, module=module)

    match conf:
        case str() | pathlib.Path():
            return Glob(_fmt(str(conf)))
        case qik.conf.GlobDep():
            return Glob(_fmt(conf.pattern))
        case qik.conf.DistDep():
            return Dist(_fmt(conf.name))
        case qik.conf.ValDep():
            return Val(_fmt(conf.key), file=_fmt(conf.file))
        case qik.conf.ModuleDep():
            return Module(_fmt(conf.name))
        case qik.conf.CmdDep():
            return Cmd(val=_fmt(conf.name), strict=conf.strict, isolated=conf.isolated)
        case qik.conf.ConstDep():
            return Const(val=_fmt(conf.val))
        case other:
            raise AssertionError(f'Unexpected dep cls "{other}"')


class Glob(BaseDep, frozen=True):
    """A dependent glob pattern"""

    @functools.cached_property
    def globs(self) -> list[str]:
        return [self.val]


class Val(BaseDep, frozen=True):
    """A value from a file."""

    file: str

    @property
    def vals(self) -> list[str]:
        # This dependency is still experimental and only used by qik graph.
        raise NotImplementedError

    @functools.cached_property
    def watch(self) -> list[str]:
        return [self.file]


class BaseCmd(BaseDep, frozen=True):
    strict: bool = False
    isolated: bool | qik.unset.UnsetType = qik.unset.UNSET

    def get_cmd_name(self) -> str:
        raise NotImplementedError

    @functools.cached_property
    def runnables(self) -> list[Runnable]:
        """Return any runnables of the dependency."""
        return [
            Runnable(name=runnable.name, obj=runnable, strict=self.strict, isolated=self.isolated)
            for runnable in qik.cmd.load(self.get_cmd_name()).runnables.values()
        ]


class Cmd(BaseCmd, frozen=True):
    """A dependent command."""

    def get_cmd_name(self) -> str:
        return self.val


class Dist(BaseDep, frozen=True):
    """A python distribution dependency."""

    @property
    def normalized(self) -> str:
        return _normalize_dist_name(self.val)

    @functools.cached_property
    def dists(self) -> list[str]:
        return [self.val]

    @functools.cached_property
    def since(self) -> list[str]:
        venv = qik.venv.load()
        if not venv.lock_file:
            raise ValueError(
                "Must configure env lock file (venvs.default.lock-file) when using --since on dists."
            )

        return venv.lock_file


class Const(BaseDep, frozen=True):
    """A constant value."""

    @functools.cached_property
    def since(self) -> list[str]:
        return ["*qik.toml"]


class Module(BaseCmd, frozen=True):
    """A python module and its associated imports."""

    strict: ClassVar[bool] = True  # type: ignore

    def get_cmd_name(self) -> str:
        return graph_cmd.analyze_cmd_name()

    @property
    def globs(self) -> list[str]:  # type: ignore
        return [str(graph_cmd.analysis_path(self.val))]


class Load(BaseDep, frozen=True):
    """Load serialized dependencies from a file."""

    default: list[str] = []  # The default globs if the file doesn't exist

    def load(self) -> Serialized | None:
        """Get the loaded dependencies."""
        try:
            return msgspec.json.decode(pathlib.Path(self.val).read_bytes(), type=Serialized)
        except (FileNotFoundError, msgspec.DecodeError):
            return None

    @property
    def dists(self) -> list[str]:  # type: ignore
        deps = self.load()
        return deps.dists if deps else []

    @property
    def globs(self) -> list[str]:  # type: ignore
        deps = self.load()
        return deps.globs if deps else self.default


def store(
    path: pathlib.Path,
    *,
    globs: list[str] | None = None,
    dists: list[str] | None = None,
    hash: bool = True,
) -> None:
    if hash:
        hash_val = Collection(*[*(globs or []), *[Dist(val=dist) for dist in dists]]).hash()
    else:
        hash_val = None

    qik.file.write(
        path, msgspec.json.encode(Serialized(globs=globs or [], dists=dists or [], hash=hash_val))
    )


class Serialized(msgspec.Struct, frozen=True, omit_defaults=True):
    """A serialized representation of dependencies, meant to be loaded from a file."""

    globs: list[str] = []
    dists: list[str] = []
    hash: str | None = None


class Collection:
    """A filterable and hashable collection of dependencies."""

    def __init__(self, *deps: str | pathlib.Path | BaseDep, module: str | None = None):
        self._deps = [dep if isinstance(dep, BaseDep) else Glob(str(dep)) for dep in deps]
        self.module = module

    @property
    def globs(self) -> set[str]:
        return {glob for dep in self._deps for glob in dep.globs} | {
            artifact for runnable in self.runnables.values() for artifact in runnable.obj.artifacts
        }

    @functools.cached_property
    def consts(self) -> set[str]:
        return {dep.val for dep in self._deps if isinstance(dep, Const)}

    @functools.cached_property
    def watch(self) -> set[str]:
        return {glob for dep in self._deps for glob in dep.watch}

    @functools.cached_property
    def since(self) -> set[str]:
        return {glob for dep in self._deps for glob in dep.since}

    @property
    def vals(self) -> set[str]:
        return {val for dep in self._deps for val in dep.vals}

    @property
    def dists(self) -> set[str]:
        return {dist for dep in self._deps for dist in dep.dists}

    @property
    def runnables(self) -> dict[str, Runnable]:
        return {
            runnable.name: runnable
            for dep in self._deps
            for runnable in dep.runnables
            if not self.module or not runnable.obj.module or self.module == runnable.obj.module
        }

    @functools.cached_property
    def consts_hash(self) -> str:
        """Hash all consts."""
        return qik.hash.strs(*self.consts)

    def hash_vals(self) -> str:
        """Hash file values."""
        return qik.hash.strs(*self.vals)

    def hash_dists(self) -> str:
        """Hash distributions."""
        return qik.hash.dists(*self.dists)

    def hash_globs(self) -> str:
        """Hash glob pattern."""
        return qik.hash.globs(*self.globs)

    def hash(self) -> str:
        """The full hash."""
        return qik.hash.strs(
            self.consts_hash, self.hash_vals(), self.hash_globs(), self.hash_dists()
        )


@functools.cache
def project_deps() -> list[BaseDep]:
    """The base dependencies for the project."""
    proj = qik.conf.project()
    return [factory(dep) for dep in proj.deps]
